# encoding: utf-8

from xmmsclient import XmmsLoop, XmmsValueC2C
from xmmsclient.service import *
import sys

class ServiceBrowse:
	def __init__(self, service):
		self._service = service
		self._results = dict()
		self._results[service._path] = service(cb = self.introspect_cb)

	def __del__(self):
		for n, r in self._results.items():
			r.disconnect()
		self._results.clear()

	def introspect_cb(self, ns, xval):
		res = self._results.pop(ns._path, None)

		if xval.is_error():
			# Occur when the target client disconnected without replying.
			return

		cxval = XmmsValueC2C(pyval = xval)
		payload = cxval.payload
		if payload and payload.is_error():
			sys.stderr.write("! Introspection error on #%d: %s\n" % (ns._clientid, payload.value()))
			return

		for attr in dir(ns):
			if attr and attr[0] == '_':
				continue
			obj = getattr(ns, attr)
			if isinstance(obj, XmmsServiceClient):
				self._results[obj._path] = obj(cb = self.introspect_cb)

		if not len(self._results):
			self.print_ns(self._service)

	def print_ns(self, ns, prefix = '+ '):
		name = ns._path and ns._path[-1] or ("Client #%d API" % ns._clientid)
		sys.stderr.write("%s%s\n" % (prefix, name))
		prefix = '  ' + prefix
		for attr in dir(ns):
			if not attr or attr[0] == '_':
				continue
			obj = getattr(ns, attr)
			if isinstance(obj, client_method):
				sys.stderr.write("%s%s()\n" % (prefix, attr))
			elif isinstance(obj, client_broadcast):
				sys.stderr.write("%s%s [broadcast]\n" % (prefix, attr))
			elif isinstance(obj, XmmsServiceClient):
				self.print_ns(obj, prefix)
			else:
				sys.stderr.write("%s%s = %r\n" % (prefix, attr, obj))

class ScWatch(XmmsLoop):
	def __init__(self, *args, **kargs):
		super(ScWatch, self).__init__(*args, **kargs)
		self._clients = dict()

	def add_clients(self, cids, ready = False):
		if isinstance(cids, int):
			cids = (cids,)
		for i in cids:
			if i == self.client_id:
				continue
			if i in self._clients:
				c, b = self._clients[i]
			else:
				c = XmmsServiceClient(self, i)
				b = None
			if ready:
				sys.stderr.write("#%02d ready\n" % i)
				b = ServiceBrowse(c)
			else:
				sys.stderr.write("#%02d connected\n" % i)
			self._clients[i] = (c, b)

	def del_clients(self, cids):
		if isinstance(cids, int):
			cids = (cids,)
		for i in cids:
			if i in self._clients:
				sys.stderr.write("#%02d disconnected\n" % i)
				del self._clients[i]

	def watch(self):
		r1 = self.broadcast_c2c_client_connected(cb = self.connected_cb)
		r2 = self.broadcast_c2c_client_disconnected(cb = self.disconnected_cb)
		r3 = self.broadcast_c2c_ready(cb = self.ready_cb)
		self.c2c_get_ready_clients(cb = self.ready_cb)
		try:
			self.loop()
		except KeyboardInterrupt:
			pass
		finally:
			r1.disconnect()
			r2.disconnect()
			r3.disconnect()

	def check_error(self, value):
		if value and value.is_error():
			sys.stderr.write("! %s\n" % value.get_error())
			return True
		return False

	def connected_cb(self, value):
		if self.check_error(value):
			return
		self.add_clients(value.value())

	def ready_cb(self, value):
		if self.check_error(value):
			return
		self.add_clients(value.value(), True)

	def disconnected_cb(self, value):
		if self.check_error(value):
			return
		self.del_clients(value.value())


if __name__ == '__main__':
	x = ScWatch(clientname = "ScWatch")
	x.connect()
	sys.stderr.write("Watching clients connections (local id: %d)\n" % x.client_id)
	sys.stderr.write("Ctrl+C to exit...\n")
	x.watch()




import xmmsapi

class XmmsSync(xmmsapi.XmmsProxy):
	"""
	A wrapper for the `xmmsclient.Xmms` class which simplifies synchronous
	communication with the XMMS2 daemon.

	Instances of this class may be used just like regular `xmmsclient.Xmms`
	objects, except that instead of returning an `XmmsResult` instance, the
	value associated with the result is returned.  If the `XmmsResult`
	indicates an error, an `XmmsError` is raised instead of returning the
	value.

	>>> import xmmsclient
	>>> xc = xmmsclient.XmmsSync('clientname')
	>>> xc.connect()
	>>> xc.playback_start()
	"""
	def __init__(self, clientname=None, xmms=None):
		"""
		This constructor takes two optional arguments. If xmms is omitted
		it will create a new underlying XMMS class otherwise it will use
		the one supplied. Clientname is the name of the client and will
		default to "Unnamed Python Client"
		"""
		if xmms is None:
			xmms = xmmsapi.Xmms(clientname)
		super(XmmsSync, self).__init__(xmms)

	def __getattr__(self, name):
		try:
			return super(XmmsSync, self).__getattr__(name)
		except AttributeError:
			pass

		attr = getattr(self._xmms, name)
		if hasattr(attr, '__call__'):
			def _(*args, **kwargs):
				ret = attr(*args, **kwargs)
				if isinstance(ret, xmmsapi.XmmsResult):
					ret.wait()
					if ret.is_error():
						raise ret.xvalue.get_error()
					return ret.value()
				return ret
			try:
				_.__doc__ = attr.__doc__
				_.func_name = '<sync version of %s>' % name
			except:
				pass
			return _
		else:
			return attr

	def __dir__(self):
		return dir(self._xmms)

from xmmsvalue import XmmsError



# encoding: utf-8
"""

 A GLib connector for PyGTK and Gobject introspection
 Tobias Rundstrom <tru@xmms.org>
 Raphaël Bois <virtualdust@gmail.com>

 GLibConnectorBase(xmms, gobject): create a glib connector using the provided
                                   gobject module.
 GLibConnector(xmms): create a glib connector using pygtk
 GiGLibConnector(xmms): create a glib connector using gobject introspection

"""
class GLibConnectorBase(object):
	def __init__(self, xmms, gobject):
		self.go = gobject
		self.in_id = None
		self.out_id = None
		self.reconnect(xmms)

	def need_out(self, i):
		if self.xmms.want_ioout() and self.out_id is None:
			self.out_id = self.go.io_add_watch(self.xmms.get_fd(), self.go.IO_OUT, self.handle_out)

	def handle_in(self, source, cond):
		if cond == self.go.IO_IN:
			return self.xmms.ioin()
		return True

	def handle_out(self, source, cond):
		if cond == self.go.IO_OUT and self.xmms.want_ioout():
			self.xmms.ioout()
		if not self.xmms.want_ioout():
			self.out_id = None
		return not self.out_id is None

	def reconnect(self, xmms=None):
		self.disconnect()
		if not xmms is None:
			self.xmms = xmms
		self.xmms.set_need_out_fun(self.need_out)
		self.in_id = self.go.io_add_watch(self.xmms.get_fd(), self.go.IO_IN, self.handle_in)
		self.out_id = None

	def disconnect(self):
		if not self.in_id is None:
			self.go.source_remove(self.in_id)
			self.in_id = None
		if not self.out_id is None:
			self.go.source_remove(self.out_id)
			self.out_id = None


class GLibConnector(GLibConnectorBase):
	def __init__(self, xmms):
		import gobject
		super(GLibConnector, self).__init__(xmms, gobject)


class GiGLibConnector(GLibConnectorBase):
	def __init__(self, xmms):
		from gi.repository import GObject as gobject
		super(GiGLibConnector, self).__init__(xmms, gobject)




from xmmsvalue import VALUE_TYPE_NONE
from xmmsvalue import VALUE_TYPE_ERROR
from xmmsvalue import VALUE_TYPE_INT64 as VALUE_TYPE_INT32
from xmmsvalue import VALUE_TYPE_INT64
from xmmsvalue import VALUE_TYPE_STRING
from xmmsvalue import VALUE_TYPE_COLL
from xmmsvalue import VALUE_TYPE_BIN
from xmmsvalue import VALUE_TYPE_LIST
from xmmsvalue import VALUE_TYPE_DICT

from xmmsvalue import COLLECTION_TYPE_REFERENCE
from xmmsvalue import COLLECTION_TYPE_UNIVERSE
from xmmsvalue import COLLECTION_TYPE_UNION
from xmmsvalue import COLLECTION_TYPE_INTERSECTION
from xmmsvalue import COLLECTION_TYPE_COMPLEMENT
from xmmsvalue import COLLECTION_TYPE_HAS
from xmmsvalue import COLLECTION_TYPE_MATCH
from xmmsvalue import COLLECTION_TYPE_TOKEN
from xmmsvalue import COLLECTION_TYPE_EQUALS
from xmmsvalue import COLLECTION_TYPE_NOTEQUAL
from xmmsvalue import COLLECTION_TYPE_SMALLER
from xmmsvalue import COLLECTION_TYPE_SMALLEREQ
from xmmsvalue import COLLECTION_TYPE_GREATER
from xmmsvalue import COLLECTION_TYPE_GREATEREQ
from xmmsvalue import COLLECTION_TYPE_ORDER
from xmmsvalue import COLLECTION_TYPE_LIMIT
from xmmsvalue import COLLECTION_TYPE_MEDIASET
from xmmsvalue import COLLECTION_TYPE_IDLIST


from xmmsapi import PLAYBACK_STATUS_STOP
from xmmsapi import PLAYBACK_STATUS_PLAY
from xmmsapi import PLAYBACK_STATUS_PAUSE

from xmmsapi import PLAYLIST_CHANGED_ADD
from xmmsapi import PLAYLIST_CHANGED_INSERT
from xmmsapi import PLAYLIST_CHANGED_SHUFFLE
from xmmsapi import PLAYLIST_CHANGED_REMOVE
from xmmsapi import PLAYLIST_CHANGED_CLEAR
from xmmsapi import PLAYLIST_CHANGED_MOVE
from xmmsapi import PLAYLIST_CHANGED_SORT
from xmmsapi import PLAYLIST_CHANGED_UPDATE

from xmmsapi import PLUGIN_TYPE_ALL
from xmmsapi import PLUGIN_TYPE_XFORM
from xmmsapi import PLUGIN_TYPE_OUTPUT

from xmmsapi import MEDIALIB_ENTRY_STATUS_NEW
from xmmsapi import MEDIALIB_ENTRY_STATUS_OK
from xmmsapi import MEDIALIB_ENTRY_STATUS_RESOLVING
from xmmsapi import MEDIALIB_ENTRY_STATUS_NOT_AVAILABLE
from xmmsapi import MEDIALIB_ENTRY_STATUS_REHASH

from xmmsapi import COLLECTION_CHANGED_ADD
from xmmsapi import COLLECTION_CHANGED_UPDATE
from xmmsapi import COLLECTION_CHANGED_RENAME
from xmmsapi import COLLECTION_CHANGED_REMOVE

from xmmsapi import PLAYBACK_SEEK_SET
from xmmsapi import PLAYBACK_SEEK_CUR

from xmmsapi import COLLECTION_NS_COLLECTIONS
from xmmsapi import COLLECTION_NS_PLAYLISTS
from xmmsapi import COLLECTION_NS_ALL



##

# Do some magic in order to make xmmsapi to see xmmsvalue
import sys, os
# XXX Do not prepend the directory. The threading import symbols from a 'collections'
# module that is not ours !!!
sys.path.append(os.path.dirname(__file__))

import xmmsapi, xmmsvalue
from xmmsapi import Xmms, XmmsLoop, XmmsResult, userconfdir_get
from xmmsvalue import XmmsValue, XmmsValueC2C
from xmmsvalue import coll_parse
from sync import XmmsSync, XmmsError
from propdict import PropDict
from consts import *

# Compatibility (deprecated)
XMMS = XmmsLoop
XMMSResult = XmmsResult
XMMSValue = XmmsValue
XMMSSync = XmmsSync
XMMSError = XmmsError

# Now, it is safe to remove the directory from the path.
sys.path.remove(os.path.dirname(__file__))
del sys, os
